#!/usr/bin/perl

package FXDB;

use strict;
use DBI;
use FXUT qw(fxmail fxlog fxabort);

my $inXact = 0;
my $DEBUG = 0;
my $dbhandle;

sub new 
{
  my ($pkg) = @_;
  my $db = bless {}, $pkg;
  
  return $db;
}


sub connect 
{
  my ($pkg, $database, $login, $passwd) = @_;
  if (!defined($passwd)) {
    $passwd = "";
    }

  ($pkg->{dbhandle} = DBI->connect("dbi:Sqlflex:$database", '', '', { AutoCommit => 1} ))
  or fxabort("Couldn't open $database $login\n");
  $dbhandle = $pkg->{dbhandle};
  $dbhandle->{PrintError} = 1;
  $dbhandle->{RaiseError} = 1;
  $dbhandle->{ChopBlanks} = 1;
  return ($dbhandle);
}

sub disconnect 
{
  my ($pkg) = @_;
  $pkg->{dbhandle}->disconnect 
      or error("couldn't close connection to db.");
  return;
}

sub xactStatus
{
  return ($inXact);
}

sub begin_xact 
{
  my ($pkg) = @_;
  $inXact = 1;
}

sub end_xact 
{
  my ($pkg) = @_;
  return if !$inXact;
  $pkg->{dbhandle}->commit
    or fxabort("couldn't exec commit\n");
  $inXact = 0;
}

sub abort_xact 
{
  my ($pkg) = @_;
  $pkg->{dbhandle}->rollback
    or fxabort("couldn't exec rollback\n");
  $inXact = 0;
}

sub escapeQuotes 
{
  my ($pkg, $q) = @_;

  if (!ref($pkg)) {
    $q = $pkg;
  }

  $q =~ s/\'/\'\'/g;
  return $q;
}

sub select 
{
  my ($pkg, $sql, $values, $benchmark) = @_;
  print "$sql\n" if ($DEBUG);

  # Prepare for benchmarking
  my $action = "";
  if ($sql =~ /^select.+(from.+)where/) {
    $action .= $1;
  } elsif ($sql =~ /^select.+(from.+);\s*/) {
    $action .= $1; 
  }

  my $sth = $pkg->{dbhandle}->prepare($sql) or
    fxabort("couldn't prepare query: $DBI::errstr: $sql\n");

  $sth->execute or
    fxabort("couldn't execute query: $DBI::errstr: $sql\n");


  my @row = ();
  while (@row = $sth->fetchrow_array() ) {
    foreach my $col (@row) { 
      if (defined($col)) {
        # $col =~ s/^(.*?)\s*$/$1/; # remove trailing spaces
        $col =~ s/^0.00000E\+00$/0/; # if exponent version of 0 then replace w/ 0 
        push (@$values, $col);
        }
      else {
        push (@$values, "NULL");
        }
      }
    }
  if ($DBI::err) {
    fxabort("($sql) didn't return proper result ($DBI::errstr).");
    }
  return @$values;
}

sub execDML
{
  my ($pkg, $command) = @_;
  print "$command\n" if ($DEBUG);

  # Parse the command to find out what kind of sql it is
  $command =~ s/^\s+//;  # remove leading white space
  if ($command !~ /^\w+\s+\w+/) {
    fxabort("Invalid command ($command).", 255);
  }

  my $action;
  if ($command =~ /^(insert into \S+)\s+/) {
    $action = $1;
  } elsif ($command =~ /^(update \S+)\s+/) {
    $action = $1;
  } elsif ($command =~ /^(delete from \S+)/) {
    $action = $1;
  } elsif ($command =~ /^(set \S+)/) {
    $action = $1;
  } elsif ($command =~ /^(info \S+)/) {
    $action = $1;
  } else {
    fxabort("Invalid command ($command).", 255);
  }

  if ($action !~ /delete/ && $action !~ /insert/ && $action !~ /update/ && $action !~ /set/ && $action !~ /info/) {
    my $errormsg = "Won't execute command ($command) action is not MI_DML";
    fxabort($errormsg, 255);
  }

  my $row_count = $pkg->{dbhandle}->do($command)
    or fxabort("couldn't carry out \"$command\"", 300);

  if ($action =~ /update/) { 
    if ($row_count < 1) {
      my $errormsg = "$action failed to affect any rows. ($command)\n";
      fxabort($errormsg, 300);
    }
  }

  # want program to act as if AutoCommit is ON when outside 
  # a transaction.
  # if ($inXact == 0 &&
  #   ($action =~ /update/ || 
  #    $action =~ /delete/ || 
  #    $action =~ /insert/) ) { 
  #   $pkg->{dbhandle}->commit
  #     or fxabort("couldn't exec commit\n");
  #  }

  return $row_count;
} # end of sub execDML

# returns 0 if value does not exist (so you can ask:  if (&exists) ...)
# returns 1 (or the number of rows containing value) if value exists
sub exists
{
  my ($pkg, $table, $column, $value, $restriction) = @_;

  my $command = "";
  if (($restriction ne "" ) && ($restriction !~ /^\s*and/)) {
    my $status = "$::this_scripts_name:  Restriction phrase ($restriction) must begin with 'and'.\n";
    fxabort($status)
  }
    
  $command = "select count($column) from $table where $column ";
  # use like if there's a "_" or "%" in $value
  if ($value =~ /[%_a-zA-Z]/) {
    $command .= "like '$value'";
  } else {
    $command .= "= '$value'";
  }
  $command .= " $restriction" if ($restriction);
  $command .= ";";

  my $row_count = &select_value ($pkg, $command);

  return $row_count;
} # end of sub exists

sub select_value
{
  my ($pkg, $sql, $benchmark) = @_;

  my @vals = ();
  &select ($pkg, $sql, \@vals, $benchmark);

  my $value = shift (@vals);

  return $value;
}


=head2 SQLinsert
   This routine is for inserting a row in the database.
   Note that text fields may also be inserted along
   with non-text fields, thereby, enabling a single insert
   statement to accomplish both.
   The arguments are
       1) Tablename. 
       2) Hash reference of fieldname(s) => value(s)

=cut

sub SQLinsert
{
  my ($pkg, $table, $col_hash_ref) = (@_);
  my %col_hash = %$col_hash_ref;

  # The only way to insert blob/text fields  is via '?' placeholders.
  my @collist = ();
  my @valuelist = ();
  my @questionlist = ();
  foreach my $colname (keys %col_hash) {
    my $value = $col_hash{$colname};
    next if ($pkg->isnull($value));
    push (@collist, $colname);
    push (@valuelist, $value);
    push (@questionlist, '?');
    }

  my $sql = "insert into $table (";
  $sql .= join(', ', @collist);
  $sql .= ") values (";
  $sql .= join(', ', @questionlist);
  $sql .= ");";

  #print "$sql\n";
  #print (join ', ', @valuelist), "\n";

  my $row_count = $pkg->{dbhandle}->do($sql, undef, @valuelist)
    or fxabort("couldn't carry out insert: \"$sql\" (" . (join ', ', @valuelist) . ")", 301);

  if ( $inXact == 0) {
    $pkg->{dbhandle}->commit
      or fxabort("couldn't exec commit\n");
    }

  return(0);
}


=head2 SQLupdate
   This routine is for updating a row in the database.
   The arguments are
       1) Tablename. 
       2) Hash reference of where fieldname(s) and 
          there corresponding value(s).
       3) Hash reference of where fieldname(s) and 
          there corresponding value(s).


   Example:
       SQLupdate("tablename", 
                  {field1=>'this is a text field',
                   field2=>'name',
                   field3=>'this is a text field',
                   field4=>'null'});
                  {wherefield1=>'wherevalue1', 
                   wherefield2=>'wherevalue2'}, 
                  [field1, field3]);
=cut

sub SQLupdate
{
  my ($pkg, $table, $col_hash_ref, $where_hash_ref) = (@_);
  my %col_hash = %$col_hash_ref;

  # build update clause for non-text fields
  my $sqlupdate = "update $table set ";
  my $count = 0;
  foreach my $colname (keys %col_hash ) {
    my $value = $col_hash{$colname};
    $sqlupdate .= ", " if ($count);
    $sqlupdate .= "$colname = ";
    if ($pkg->isnull($value)) {
      $sqlupdate .= "null ";
      }
    elsif ($value eq 'today') {
      $sqlupdate .= "today ";
      }
    else {
      $sqlupdate .= "'$value' ";
      }
    $count++;
    }

  my $where_clause = $pkg->getwhereclause($where_hash_ref);
  if ($where_clause eq "") {
    fxabort("update must have a where clause !!!", 302);
    }
  $sqlupdate .= $where_clause;

  $pkg->execDML($sqlupdate);

  return(0);
}


=head2 SQLdelete
   This routine is for deleting rows in the database.
   The arguments are
       1) Tablename. 
       2) Hash reference of where fieldname(s) and 
          there corresponding value(s).


   Example:
       SQLdelete("tablename", 
                  {wherefield1=>'wherevalue1', 
                   wherefield2=>'wherevalue2'};
=cut

sub SQLdelete
{
  my ($pkg, $table, $where_hash_ref) = (@_);

  my $sql = "delete from $table ";
  my $where_clause = $pkg->getwhereclause($where_hash_ref);
  if ($where_clause eq "") {
    fxabort("delete must have a where clause !!!", 302);
    }
  $sql .= $where_clause;
  $pkg->execDML($sql);
  return(0);
}


sub SQLselectall_arrayref
{
  my ($pkg, $sql) = (@_);
  my $array_ref = $pkg->{dbhandle}->selectall_arrayref($sql) or
    fxabort("couldn't query (selectall_arrayref): $DBI::errstr: $sql\n");
  return($array_ref);
}

sub SQLprepare
{
  my ($pkg, $sql) = (@_);
  my $sth = $pkg->{dbhandle}->prepare($sql) or
    fxabort("couldn't prepare: $DBI::errstr: $sql\n");
  $pkg->{sth} = $sth;
  return($pkg);
}

sub SQLexecute
{
  my ($pkg, @args) = (@_);
  $pkg->{sth}->execute(@args) or
    fxabort("couldn't execute: $DBI::errstr: @args\n");
  return();
}

sub SQLfetchrow
{
  my ($pkg) = (@_);
  my @rows = $pkg->{sth}->fetchrow(); 

  # returns 0 is OK
  # fxabort("couldn't execute: $DBI::errstr: @args\n");
  return(@rows);
}


sub getwhereclause
{
  my $pkg = $_[0];
  my $where_hash_ref = $_[1];
  my %where_hash = %$where_hash_ref;

  # build where clause
  my $where_clause = " where ";
  my $where_count = 0;
  foreach my $colname (keys %where_hash ) {
    my $value = $where_hash{$colname};
    $where_clause .= "and " if ($where_count);
    $where_clause .= "$colname = ";
    if ($pkg->isnull($value)) {
      $where_clause .= "null ";
      }
    elsif ($value eq 'today') {
      $where_clause .= "today ";
      }
    else {
      $where_clause .= "'$value' ";
      }
    $where_count++;
    }

  if ($where_count == 0) {
    return("");
    }
  return($where_clause);
}


sub comma_quote_list
{
  my $pkg = shift(@_);
  my $list = join("', '", @_);
  $list = "'" . $list;
  $list .= "'";
  return($list);
}

sub isnull
{
  my ($pkg, $value) = (@_);

  if (!defined($value) || $value eq "" || $value eq "null" || $value eq "NULL") {
    return(1);
    }

  return(0);
}


1;
